#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['idf_build_apps',
 'idf_build_apps.junit',
 'idf_build_apps.manifest',
 'idf_build_apps.yaml']

package_data = \
{'': ['*']}

install_requires = \
['pyparsing', 'pyyaml', 'packaging', 'pydantic~=2.0', 'argcomplete>=3']

extras_require = \
{":python_version < '3.11'": ['toml'],
 "dev:python_version < '3.8'": ['typing-extensions'],
 'doc': ['sphinx',
         'sphinx-rtd-theme',
         'sphinx_copybutton',
         'myst-parser',
         'sphinxcontrib-mermaid',
         'sphinx-argparse'],
 'test': ['pytest', 'pytest-cov']}

entry_points = \
{'console_scripts': ['idf-build-apps = idf_build_apps:main.main']}

setup(name='idf-build-apps',
      version='2.4.2',
      description='Tools for building ESP-IDF related apps.',
      author=None,
      author_email='Fu Hanxi <fuhanxi@espressif.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.7',
     )
